# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
from abc import ABC, abstractmethod
import numpy as np
from typing import Callable, List, Any
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.core.common import Graph, Logger
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.gptq.common.gptq_graph import get_compare_points
from model_compression_toolkit.core.common.model_builder_mode import ModelBuilderMode


class GPTQTrainer(ABC):
    """
    Abstract GPTQ training class for fine-tuning a quantized model
    """

    def __init__(self,
                 graph_float: Graph,
                 graph_quant: Graph,
                 gptq_config: GradientPTQConfig,
                 fw_impl: FrameworkImplementation,
                 fw_info: FrameworkInfo):
        """
        Build two models from a graph: A teacher network (float model) and a student network (quantized model).
        Use the dataset generator to pass images through the teacher and student networks to get intermediate
        layers outputs. Use the outputs to compute the observed loss and to back-propagate the error
        in the student network, to minimize it in the next similar steps.
        All parameters (such as number of iterations, optimizer, etc.) are in GradientPTQConfig.
        Args:
            graph_float: Graph to build a float networks from.
            graph_quant: Graph to build a quantized networks from.
            gptq_config: GradientPTQConfig with parameters about the tuning process.
            fw_impl: Framework implementation
            fw_info: Framework information
        """
        self.graph_float = copy.deepcopy(graph_float)
        self.graph_quant = copy.deepcopy(graph_quant)
        self.gptq_config = gptq_config
        self.fw_impl = fw_impl
        self.fw_info = fw_info

        # ----------------------------------------------
        # Build two models and create compare nodes
        # ----------------------------------------------
        self.compare_points, _, self.compare_points_mean, self.compare_points_std = get_compare_points(self.graph_float)

        self.float_model, self.float_user_info = fw_impl.model_builder(self.graph_float,
                                                                       mode=ModelBuilderMode.FLOAT,
                                                                       append2output=self.compare_points,
                                                                       fw_info=self.fw_info)

        self.fxp_model, self.gptq_user_info = self.build_gptq_model()

    def get_optimizer_with_param(self,
                                 flattened_trainable_weights: List[Any],
                                 flattened_bias_weights: List[Any],
                                 trainable_quantization_parameters: List[Any],
                                 temperature_weights: List[Any]) -> List[Any]:
        """
        Create Optimizers with their trainable parameters
        Args:
            flattened_trainable_weights: list of trainable weights parameters (flattened)
            flattened_bias_weights: list of trainable bias parameters (flattened)
            trainable_quantization_parameters: list of trainable quantization parameters
            temperature_weights: list of temperature weights variables
        Returns:
            List of Optimizer objects with parameters
        """

        w2train = [*flattened_trainable_weights]
        if self.gptq_config.is_gumbel:
            if self.gptq_config.quantizer_config.temperature_learning:
                w2train.extend(temperature_weights)

        optimizer_with_param = [(self.gptq_config.optimizer, w2train)]
        if self.gptq_config.train_bias or self.gptq_config.quantization_parameters_learning:
            w2train_res = []
            if self.gptq_config.train_bias:
                if self.gptq_config.optimizer_bias is not None:
                    optimizer_with_param.append((self.gptq_config.optimizer_bias, flattened_bias_weights))
                else:
                    w2train_res.extend(flattened_bias_weights)
                    if self.gptq_config.optimizer_rest is None:
                        Logger.error(
                            "To enable bias micro training an additional optimizer is required, please define the optimizer_rest")
            if self.gptq_config.quantization_parameters_learning:
                if self.gptq_config.optimizer_quantization_parameter is not None:  # Ability to override optimizer
                    optimizer_with_param.append((self.gptq_config.optimizer_quantization_parameter,
                                                 trainable_quantization_parameters))
                else:
                    w2train_res.extend(trainable_quantization_parameters)
                if self.gptq_config.optimizer_rest is None:
                    Logger.error(
                        "To enable bias micro training an additional optimizer is required, please define the optimizer_rest")
            optimizer_with_param.append((self.gptq_config.optimizer_rest, w2train_res))

        return optimizer_with_param


    def compute_jacobian_based_weights(self,
                                       representative_data_gen: Callable) -> np.ndarray:
        """
        Computes the jacobian-based weights using the framework's model_grad method per batch of images.

        Args:
            representative_data_gen: Dataset used for inference to compute the jacobian-based weights.

        Returns: A vector of weights, one for each compare point,
        to be used for the loss metric weighted average computation when running GPTQ training.
        """
        if self.gptq_config.use_jac_based_weights:
            images = self._generate_images_batch(representative_data_gen, self.gptq_config.num_samples_for_loss)
            points_apprx_jacobians_weights = []
            for i in range(1, images.shape[0] + 1):
                # Note that in GPTQ loss weights computation we assume that there aren't replacement output nodes,
                # therefore, output_list is just the graph outputs, and we don't need the tuning factor for
                # defining the output weights (since the output layer is not a compare point).
                image_ip_gradients = self.fw_impl.model_grad(self.graph_float,
                                                             {inode: self.fw_impl.to_tensor(images[i - 1:i]) for inode in
                                                              self.graph_float.get_inputs()},
                                                             self.compare_points,
                                                             output_list=[n.node for n in
                                                                          self.graph_float.get_outputs()],
                                                             all_outputs_indices=[],
                                                             alpha=0,
                                                             norm_weights=self.gptq_config.norm_weights)
                points_apprx_jacobians_weights.append(image_ip_gradients)
            return np.mean(points_apprx_jacobians_weights, axis=0)
        else:
            num_nodes = len(self.compare_points)
            return np.asarray([1 / num_nodes for _ in range(num_nodes)])

    @staticmethod
    def _generate_images_batch(representative_data_gen: Callable, num_samples_for_loss: int) -> np.ndarray:
        """
        Construct batches of image samples for inference.

        Args:
            representative_data_gen: A callable method to retrieve images from Dataset.
            num_samples_for_loss: Num of total images for evaluation.

        Returns: A tensor of images batches
        """
        # First, select images to use for all measurements.
        samples_count = 0  # Number of images we used so far to compute the distance matrix.
        images = []
        while samples_count < num_samples_for_loss:
            # Get a batch of images to infer in both models.
            inference_batch_input = representative_data_gen()
            num_images = inference_batch_input[0].shape[0]

            # If we sampled more images than we should,
            # we take only a subset of these images and use only them.
            if num_images > num_samples_for_loss - samples_count:
                inference_batch_input = [x[:num_samples_for_loss - samples_count] for x in inference_batch_input]
                assert num_samples_for_loss - samples_count == inference_batch_input[0].shape[0]
                num_images = num_samples_for_loss - samples_count

            images.append(inference_batch_input[0])
            samples_count += num_images
        return np.concatenate(images, axis=0)


    @abstractmethod
    def build_gptq_model(self):
        """
        Build the GPTQ model with QuantizationWrappers
        Returns:
            Quantized graph for GPTQ fine-tuning, GPTQ graph user info
        """
        raise NotImplemented(f'{self.__class__.__name__} have to implement the '
                             f'framework\'s GPTQ model builder method.')

    @abstractmethod
    def train(self, representative_data_gen: Callable):
        """
        Train the quantized model using GPTQ training process
        Args:
            representative_data_gen: Dataset to use for inputs of the models.
        """
        raise NotImplemented(f'{self.__class__.__name__} have to implement the '
                             f'framework\'s train method.')

    @abstractmethod
    def update_graph(self) -> Graph:
        """
        Update a graph using GPTQ after minimizing the loss between the float model's output
        and the quantized model's outputs.
        Returns:
            Updated graph after GPTQ.
        """
        raise NotImplemented(f'{self.__class__.__name__} have to implement the '
                             f'framework\'s update_graph method.')


def gptq_training(graph_float: Graph,
                  graph_quant: Graph,
                  gptq_config: GradientPTQConfig,
                  representative_data_gen: Callable,
                  fw_impl: FrameworkImplementation,
                  fw_info: FrameworkInfo) -> Graph:
    """
    GPTQ training process using knowledge distillation with a teacher network (float model) and a student network (quantized model).
    Args:
        graph_float: Graph to build a float networks from.
        graph_quant: Graph to build a quantized networks from.
        gptq_config: GradientPTQConfig with parameters about the tuning process.
        representative_data_gen: Dataset to use for inputs of the models.
        fw_impl: Framework implementation
        fw_info: Framework information

    Returns:
        Quantized graph for export

    """
    # Get GPTQ object and initialize it
    gptq_trainer_obj = fw_impl.get_gptq_trainer_obj()

    gptq_trainer = gptq_trainer_obj(graph_float,
                                    graph_quant,
                                    gptq_config,
                                    fw_impl,
                                    fw_info,
                                    representative_data_gen)

    # Training process
    gptq_trainer.train(representative_data_gen)

    # Update graph
    graph_quant = gptq_trainer.update_graph()

    return graph_quant



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np

from typing import List, Callable

from model_compression_toolkit.core.common import BaseNode, Logger
from model_compression_toolkit.core.common.constants import THRESHOLD, SIGNED, RANGE_MIN, RANGE_MAX
from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import calculate_delta
from model_compression_toolkit.core.common.target_platform import QuantizationMethod

# Supporting other quantizer types in the future
from model_compression_toolkit.exporter.fully_quantized.pytorch.quantizers.fq_quantizer import FakeQuantQuantizer
from model_compression_toolkit.exporter.fully_quantized.pytorch.quantizers.uniform_weights_quantizer import \
    UniformWeightsQuantizer
import torch

SUPPORTED_WEIGHT_QUANTIZER_TYPES = [QuantizationMethod.POWER_OF_TWO,
                                    QuantizationMethod.SYMMETRIC,
                                    QuantizationMethod.UNIFORM]

SUPPORTED_ACTIVATION_QUANTIZER_TYPES = [QuantizationMethod.POWER_OF_TWO,
                                        QuantizationMethod.SYMMETRIC,
                                        QuantizationMethod.UNIFORM]


def get_weights_quantizer_for_node(node: BaseNode) -> List[Callable]:
    """
    Get weights quantizer for a node.

    Args:
        node: Node to create a weight quantizer for.

    Returns:
        Quantizer for the node's weights.

    """
    if node.final_weights_quantization_cfg is None:
        Logger.critical(f'Can not set quantizer for a node with no final weights quantization configuration')

    node_w_qc = node.final_weights_quantization_cfg
    weights_quantization_method = node_w_qc.weights_quantization_method

    if weights_quantization_method not in SUPPORTED_WEIGHT_QUANTIZER_TYPES:
        Logger.error(f'Fully quantized models are now supported for {SUPPORTED_WEIGHT_QUANTIZER_TYPES} quantization methods, but node has {weights_quantization_method} quantization method')

    # Compute quantizer params based on node's quantization params
    if weights_quantization_method in [QuantizationMethod.POWER_OF_TWO, QuantizationMethod.SYMMETRIC]:
        weight_thresholds = node_w_qc.weights_quantization_params.get(THRESHOLD)
        assert weight_thresholds is not None
        if weights_quantization_method == QuantizationMethod.POWER_OF_TWO:
            is_threshold_pot = np.all([int(np.log2(x)) == np.log2(x) for x in weight_thresholds.flatten()])
            if not is_threshold_pot:
                Logger.error(f'Expected threshold to be power of 2 but is {weight_thresholds}')

        min_range = -weight_thresholds
        max_range = weight_thresholds - calculate_delta(weight_thresholds,
                                                        n_bits=node_w_qc.weights_n_bits,
                                                        signed=True)

    else:
        Logger.error(f'For now fully quantized models support only {SUPPORTED_WEIGHT_QUANTIZER_TYPES} for weights quantization, but found {weights_quantization_method}')

    return [UniformWeightsQuantizer(num_bits=node_w_qc.weights_n_bits,
                                    max_range=max_range,
                                    min_range=min_range,
                                    quantization_method=node_w_qc.weights_quantization_method,
                                    per_channel=node_w_qc.weights_per_channel_threshold,
                                    output_channels_axis=node_w_qc.weights_channels_axis
                                    )]


def get_activations_quantizer_for_node(node: BaseNode) -> List[Callable]:
    """
    Get activation quantizer for a node.

    Args:
        node: Node to create an activation quantizer for.

    Returns:
        Quantizer for the node's activations.

    """

    if node.final_activation_quantization_cfg is None:
        Logger.critical(f'Can not set quantizer for a node with no final activation quantization configuration')

    node_act_qc = node.final_activation_quantization_cfg
    activation_quantization_method = node_act_qc.activation_quantization_method

    if activation_quantization_method not in SUPPORTED_ACTIVATION_QUANTIZER_TYPES:
        Logger.error(
            f'Fully quantized models are now supported for {SUPPORTED_ACTIVATION_QUANTIZER_TYPES} quantization methods, '
            f'but node has {activation_quantization_method} quantization method')

    activation_thresholds = node_act_qc.activation_quantization_params.get(THRESHOLD)

    if activation_quantization_method in [QuantizationMethod.POWER_OF_TWO, QuantizationMethod.SYMMETRIC]:
        if activation_quantization_method == QuantizationMethod.POWER_OF_TWO:
            is_threshold_pot = np.all([int(np.log2(x)) == np.log2(x) for x in activation_thresholds.flatten()])
            if not is_threshold_pot:
                Logger.error(f'Expected threshold to be power of 2 but is {node_act_qc.activation_quantization_params.get(THRESHOLD)}')

        min_range = 0
        if node_act_qc.activation_quantization_params.get(SIGNED):
            min_range = -activation_thresholds

        max_range = activation_thresholds - calculate_delta(
            activation_thresholds,
            n_bits=node_act_qc.activation_n_bits,
            signed=node_act_qc.activation_quantization_params.get(SIGNED))

    elif activation_quantization_method in [QuantizationMethod.UNIFORM]:
        min_range = node_act_qc.activation_quantization_params.get(RANGE_MIN)
        max_range = node_act_qc.activation_quantization_params.get(RANGE_MAX)

    else:
        Logger.error(f'For now fully quantized models support only {SUPPORTED_ACTIVATION_QUANTIZER_TYPES} for activation quantization, but found {activation_quantization_method}')

    return [FakeQuantQuantizer(nbits=node_act_qc.activation_n_bits,
                              min_range=min_range,
                              max_range=max_range,
                              quantization_method=activation_quantization_method)]



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf
import numpy as np

from model_compression_toolkit import GumbelConfig
from model_compression_toolkit.gptq.keras.quantizer import quant_utils as qutils
from model_compression_toolkit.gptq.keras.quantizer.gumbel_rounding.base_gumbel_rounding import GumbelRoundingBase
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper
from tensorflow.python.framework.tensor_shape import TensorShape
from model_compression_toolkit.core.common.defaultdict import DefaultDict
from typing import Dict, Any, List
from model_compression_toolkit.gptq.keras.quantizer.gumbel_rounding.gumbel_softmax import gumbel_softmax, ste_gumbel
from model_compression_toolkit.core.common.constants import THRESHOLD, GUMBEL_MAX_ITER
from model_compression_toolkit.gptq.common import gptq_constants


def gumbel_rounding_symmetric_quantizer(input_tensor: tf.Tensor,
                                        auxvar_tensor: tf.Variable,
                                        max_tensor: tf.Tensor,
                                        num_bits: int,
                                        signed: bool,
                                        power_of_two: bool) -> tf.Tensor:
    """
    Quantize a tensor symmetrically with maximum LSBs shift.
    Args:
        input_tensor: Tensor to quantize. values of this tensor are not changed during gptq.
        auxvar_tensor: Tensor that manifests the bit shift the weight due to gptq.
        max_tensor: Tensor with max values to compute the threshold.
        num_bits: Num of bits to use.
        signed: Signedness of the quantization range.
        power_of_two: Whether the threshold should be constrained or not.

    Returns:
        A quantized tensor.
    """

    if power_of_two:
        max_tensor = qutils.power_of_two_max(max_tensor)
    delta = qutils.calculate_delta(max_tensor, num_bits, signed)
    input_tensor = tf.stop_gradient(input_tensor)
    input_tensor_int = qutils.ste_round(input_tensor / delta)
    tensor_q = input_tensor_int + auxvar_tensor
    min_int = -int(signed) * (2 ** (num_bits - int(signed)))
    max_int = (2 ** (num_bits - int(signed))) - 1
    return delta * qutils.clip(tensor_q, max_val=max_int, min_val=min_int)

class SymmetricGumbelRounding(GumbelRoundingBase):
    """
    Trainable constrained quantizer to quantize a layer inputs.
    """
    PTQ_THRESHOLD = "_ptq_threshold"
    SCALE_PTQ = "_scale"

    def __init__(self, num_bits: int,
                 per_axis: bool,
                 signed: bool,
                 power_of_two: bool,
                 quantization_parameter_learning: bool,
                 threshold_values: np.ndarray,
                 gumbel_config: GumbelConfig,
                 quantization_axis: int = -1,
                 max_lsbs_change_map: dict = DefaultDict({}, lambda: 1),
                 max_iteration: int = GUMBEL_MAX_ITER,
                 gumbel_scale: float = 1.0):
        """
        Initialize a TrainableWeightQuantizer object with parameters to use
        for the quantization.

        Args:
            num_bits: Number of bits to use for the quantization.
            per_axis: Whether to quantize per-channel or per-tensor.
            signed: Signedness to use for the quantization range.
            threshold_values: Threshold to use for the quantization.
            gumbel_config: A class with the gumbel rounding configurations.
            quantization_axis: Axis of tensor to use for the quantization.
            power_of_two: Whether the threshold should be constrained or not.
            max_lsbs_change_map: a mapping between number of bits to max lsb change.
            max_iteration: The number of iteration of gptq.
            gumbel_scale: A normalization factor for the gumbel tensor values
        """
        super().__init__(num_bits, per_axis, signed, True, power_of_two, quantization_parameter_learning,
                         quantization_axis, gumbel_config,
                         max_lsbs_change_map,
                         max_iteration)
        self.threshold_shape = np.asarray(threshold_values).shape
        self.threshold_values = np.reshape(np.asarray(threshold_values), [-1]) if self.per_axis else float(
            threshold_values)
        self.k_threshold = len(self.threshold_values) if self.per_axis else 1
        self.gumbel_scale = gumbel_scale

    def build(self,
              tensor_shape: TensorShape,
              name: str,
              layer: QuantizeWrapper) -> Dict[str, tf.Variable]:
        """
        Add min and max variables to layer.
        Args:
            tensor_shape: Tensor shape the quantizer quantize.
            name: Prefix of variables names.
            layer: Layer to add the variables to. The variables are saved
            in the layer's scope.

        Returns:
            Dictionary of new variables.
        """
        super().build(tensor_shape, name, layer)

        ptq_threshold_tensor = layer.add_weight(
            name + self.PTQ_THRESHOLD,
            shape=self.k_threshold,
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=False)
        ptq_threshold_tensor.assign(self.threshold_values)

        self.quantizer_parameters.update({self.PTQ_THRESHOLD: ptq_threshold_tensor})

        if self.quantization_parameter_learning and not self.power_of_two:
            scale = layer.add_weight(
                name + self.SCALE_PTQ,
                shape=self.k_threshold,
                initializer=tf.keras.initializers.Constant(1.0),
                trainable=True)
            self.quantizer_parameters.update({self.SCALE_PTQ: scale})

        return self.quantizer_parameters

    def get_quantization_variable(self) -> List[tf.Tensor]:
        """
        This function return a list of quantizer parameters.
        Returns: A list of the quantizer parameters

        """
        if self.quantization_parameter_learning and not self.power_of_two:
            return [self.quantizer_parameters[self.SCALE_PTQ]]
        else:
            return [self.quantizer_parameters[self.PTQ_THRESHOLD]]

    def __call__(self, inputs: tf.Tensor,
                 training: bool,
                 weights: Dict[str, tf.Variable],
                 **kwargs: Dict[str, Any]):
        """
        Quantize a tensor.
        Args:
            inputs: Input tensor to quantize.
            training: Whether the graph is in training mode.
            weights: Dictionary of weights the quantizer can use to quantize the tensor.
            **kwargs: Additional variables the quantizer may receive.

        Returns:
            The quantized tensor.
        """

        auxvar = weights[gptq_constants.AUXVAR]
        ar_iter = weights[gptq_constants.GPTQ_ITER]
        ptq_threshold_tensor = weights[self.PTQ_THRESHOLD]
        aux_index_shift = weights[gptq_constants.AUXSHIFT]
        self.update_iteration(training, ar_iter)
        if self.per_axis:
            input_shape = inputs.shape
            n_axis = len(input_shape)
            quantization_axis = n_axis + self.quantization_axis if self.quantization_axis < 0 else \
                self.quantization_axis
            reshape_shape = [-1 if i == quantization_axis else 1 for i in range(n_axis)]

            reshape_shape_aux_ind = [-1, *[1 for _ in range(n_axis)]]
            #####################################################
            # Gumbel Softmax
            #####################################################
            if training:
                p_t = gumbel_softmax(auxvar, self.tau, self.g_t, gumbel_scale=self.gumbel_scale)
            else:
                p_t = gumbel_softmax(auxvar, self.minimal_temp, 0)
                p_t = ste_gumbel(p_t)
            self.p_t = p_t
            #####################################################
            # Calculate v hat and threshold hat
            #####################################################
            ptq_threshold_tensor_hat = tf.reshape(ptq_threshold_tensor, reshape_shape)
            auxvar_hat = tf.reduce_sum(p_t * tf.reshape(aux_index_shift, reshape_shape_aux_ind), axis=0)
            #####################################################
            # Quantized Input
            #####################################################
            q_tensor = gumbel_rounding_symmetric_quantizer(inputs, auxvar_hat,
                                                           ptq_threshold_tensor_hat,
                                                           self.num_bits,
                                                           self.signed,
                                                           self.power_of_two)
            if self.quantization_parameter_learning and not self.power_of_two:
                scale = tf.reshape(self.quantizer_parameters[self.SCALE_PTQ], reshape_shape)
                q_tensor *= scale

            return q_tensor
        else:
            return gumbel_rounding_symmetric_quantizer(inputs, auxvar,
                                                       ptq_threshold_tensor,
                                                       self.num_bits,
                                                       self.signed,
                                                       self.power_of_two)

    def get_quant_config(self, layer) -> Dict[str, np.ndarray]:
        """
        Returns the config used to edit NodeQuantizationConfig after GPTQ retraining

        Args:
            layer: quantized layer

        Returns:
            A dictionary of attributes the quantize_config retraining has changed during GPTQ retraining.
            Keys must match NodeQuantizationConfig attributes

        """

        if self.power_of_two:
            old_threshold = self.quantizer_parameters[self.PTQ_THRESHOLD]
            old_threshold = qutils.power_of_two_max(old_threshold)
        else:
            old_threshold = self.quantizer_parameters[self.PTQ_THRESHOLD]
            if self.quantization_parameter_learning:
                old_threshold = old_threshold * self.quantizer_parameters[self.SCALE_PTQ]
        old_threshold = old_threshold.numpy().reshape(self.threshold_shape)
        return {THRESHOLD: old_threshold}



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np

import model_compression_toolkit.core.common.quantization.quantization_config as qc
from model_compression_toolkit.core.common.constants import MIN_THRESHOLD, RANGE_MIN, RANGE_MAX
from model_compression_toolkit.core.common.quantization.quantization_params_generation.qparams_search import \
    qparams_uniform_selection_tensor_search, qparams_uniform_selection_histogram_search
from model_compression_toolkit.core.common.quantization.quantization_params_generation.error_functions import \
    get_threshold_selection_tensor_error_function, get_threshold_selection_histogram_error_function
from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import get_tensor_max, \
    get_tensor_min
from model_compression_toolkit.core.common.target_platform import QuantizationMethod

def uniform_selection_tensor(tensor_data: np.ndarray,
                             p: int,
                             n_bits: int,
                             per_channel: bool = False,
                             channel_axis: int = 1,
                             n_iter: int = 10,
                             min_threshold: float = MIN_THRESHOLD,
                             quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the optimal quantization range based on the provided QuantizationErrorMethod
    to uniformly quantize the tensor.
    Different search is applied, depends on the value of the selected QuantizationErrorMethod.

    Args:
        tensor_data: Tensor content as Numpy array.
        p: p-norm to use for the Lp-norm distance.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the quantization should be per-channel or not.
        channel_axis: Output channel index.
        n_iter: Number of iterations to search for the optimal threshold (not used for this method).
        min_threshold: Minimal threshold to use if threshold is too small (not used for this method).
        quant_error_method: an error function to optimize the range parameters' selection accordingly.

    Returns:
        Optimal quantization range to quantize the tensor uniformly.
    """
    tensor_min = get_tensor_min(tensor_data, per_channel, channel_axis)
    tensor_max = get_tensor_max(tensor_data, per_channel, channel_axis)

    if quant_error_method == qc.QuantizationErrorMethod.NOCLIPPING:
        mm = tensor_min, tensor_max
    else:
        error_function = get_threshold_selection_tensor_error_function(QuantizationMethod.UNIFORM, quant_error_method, p, norm=False)
        mm = qparams_uniform_selection_tensor_search(error_function,
                                                     tensor_data,
                                                     tensor_min,
                                                     tensor_max,
                                                     n_bits,
                                                     per_channel,
                                                     channel_axis)
    return {RANGE_MIN: mm[0],
            RANGE_MAX: mm[1]}


def uniform_selection_histogram(bins: np.ndarray,
                                counts: np.ndarray,
                                p: int,
                                n_bits: int,
                                min_value: float,
                                max_value: float,
                                constrained: bool = True,
                                n_iter: int = 20,
                                min_threshold: float = MIN_THRESHOLD,
                                quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the optimal quantization range based on the provided QuantizationErrorMethod
    to uniformly quantize the histogram.
    Different search is applied, depends on the value of the selected QuantizationErrorMethod.

    Args:
        bins: Bins values of the histogram.
        counts: Bins counts of the histogram.
        p: p-norm to use for the Lp-norm distance (used only for lp threshold selection).
        n_bits: Number of bits to quantize the tensor.
        min_value: Min value (not used for this method).
        max_value: Max value (not used for this method).
        constrained: Whether the threshold should be constrained or not (not used for this method).
        n_iter: Number of iteration ot search for the threshold (not used for this method).
        min_threshold: Minimal threshold to use if threshold is too small (not used for this method).
        quant_error_method: an error function to optimize the range parameters selection accordingly.

    Returns:
        Optimal quantization range to quantize the histogram uniformly.
    """
    tensor_min = np.min(bins[:-1][counts > 0])
    tensor_max = np.max(bins[1:][counts > 0])
    tensor_min_max = np.array([tensor_min, tensor_max])

    if quant_error_method == qc.QuantizationErrorMethod.NOCLIPPING:
        mm = tensor_min_max
    else:
        error_function = get_threshold_selection_histogram_error_function(QuantizationMethod.UNIFORM, quant_error_method, p)
        mm = qparams_uniform_selection_histogram_search(error_function,
                                                        tensor_min_max,
                                                        bins,
                                                        counts,
                                                        n_bits)

    return {RANGE_MIN: mm[0],
            RANGE_MAX: mm[1]}


def uniform_no_clipping_selection_min_max(bins: np.ndarray,
                                          counts: np.ndarray,
                                          p: int,
                                          n_bits: int,
                                          min_value: float,
                                          max_value: float,
                                          constrained: bool = False,
                                          n_iter: int = 20,
                                          min_threshold: float = MIN_THRESHOLD,
                                          quant_error_method: qc.QuantizationErrorMethod =
                                          qc.QuantizationErrorMethod.NOCLIPPING) -> dict:
    """
    Gets a quantization rage between min and max numbers.

    Returns:
        A constrained threshold of the min/max values.
    """
    return uniform_selection_histogram(np.asarray([min_value, max_value]),  # histogram with min-max values only
                                       np.asarray([1]),  # passing dummy counts just to make the dimensions work
                                       p,
                                       n_bits,
                                       min_value,
                                       max_value,
                                       constrained,
                                       n_iter,
                                       min_threshold=min_threshold,
                                       quant_error_method=qc.QuantizationErrorMethod.NOCLIPPING)




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
import tensorflow as tf

# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
if tf.__version__ < "2.6":
    from tensorflow.python.keras.layers import Layer
else:
    from keras.engine.base_layer import Layer

from typing import Any, Dict, List, Tuple
from tensorflow.python.util.object_identity import Reference as TFReference
from model_compression_toolkit.core.common.constants import EPS
from model_compression_toolkit.core.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import BaseNode, Graph
from model_compression_toolkit.core.common.graph.edge import EDGE_SINK_INDEX
from model_compression_toolkit.core.keras.back2framework.instance_builder import OperationHandler
from model_compression_toolkit.core.common.logger import Logger


def build_input_tensors_list(node: BaseNode,
                             graph: Graph,
                             node_to_output_tensors_dict: Dict[BaseNode, List[TFReference]]) -> List[List[TFReference]]:
    """
    Given a node, build a list of input tensors the node gets. The list is built
    based on the node's incoming edges and previous nodes' output tensors.

    Args:
        node: Node to build its input tensors list.
        graph: Graph the node is in.
        node_to_output_tensors_dict: A dictionary from a node to its output tensors.

    Returns:
        A list of the node's input tensors.
    """

    input_tensors = []
    # Go over a sorted list of the node's incoming edges, and for each source node get its output tensors.
    # Append them in a result list.
    for ie in graph.incoming_edges(node, sort_by_attr=EDGE_SINK_INDEX):
        _input_tensors = [node_to_output_tensors_dict[ie.source_node][ie.source_index]]
        input_tensors.append(_input_tensors)
    return input_tensors


def run_operation(n: BaseNode,
                  input_tensors: List[List[TFReference]],
                  op_func: Layer,
                  input_nodes_to_input_tensors: Dict[BaseNode, Any]) -> List[TFReference]:
    """
    Applying the layer (op_func) to the input tensors (input_tensors).

    Args:
        n: The corresponding node of the layer it runs.
        input_tensors: List of references to Keras tensors that are the layer's inputs.
        op_func: Layer to apply to the input tensors.
        input_nodes_to_input_tensors: A dictionary from a node to its input tensors.

    Returns:
        A list of references to Keras tensors. The layer's output tensors after applying the
        layer to the input tensors.
    """

    if len(input_tensors) == 0:  # Placeholder handling
        out_tensors_of_n = input_nodes_to_input_tensors[n]
    else:
        input_tensors = [tensor for tensor_list in input_tensors for tensor in tensor_list]  # flat list of lists
        # Build a functional node using its args
        if isinstance(n, FunctionalNode):
            if n.inputs_as_list:  # If the first argument should be a list of tensors:
                out_tensors_of_n = op_func(input_tensors, *n.op_call_args, **n.op_call_kwargs)
            else:  # If the input tensors should not be a list but iterated:
                out_tensors_of_n = op_func(*input_tensors, *n.op_call_args, **n.op_call_kwargs)
        else:
            # If operator expects a single input tensor, it cannot be a list as it should have a dtype field.
            if len(input_tensors) == 1:
                input_tensors = input_tensors[0]
            out_tensors_of_n = op_func(input_tensors)

    return out_tensors_of_n


def keras_iterative_approx_jacobian_trace(graph_float: common.Graph,
                                          model_input_tensors: Dict[BaseNode, np.ndarray],
                                          interest_points: List[BaseNode],
                                          output_list: List[BaseNode],
                                          all_outputs_indices: List[int],
                                          alpha: float = 0.3,
                                          n_iter: int = 50,
                                          norm_weights: bool = True) -> List[float]:
    """
    Computes an approximation of the power of the Jacobian trace of a Keras model's outputs with respect to the feature maps of
    the set of given interest points. It then uses the power of the Jacobian trace for each interest point and normalized the
    values, to be used as weights for weighted average in distance metric computation.

    Args:
        graph_float: Graph to build its corresponding Keras model.
        model_input_tensors: A mapping between model input nodes to an input batch.
        interest_points: List of nodes which we want to get their feature map as output, to calculate distance metric.
        output_list: List of nodes that considered as model's output for the purpose of gradients computation.
        all_outputs_indices: Indices of the model outputs and outputs replacements (if exists),
            in a topological sorted interest points list.
        alpha: A tuning parameter to allow calibration between the contribution of the output feature maps returned
            weights and the other feature maps weights (since the gradient of the output layers does not provide a
            compatible weight for the distance metric computation).
        n_iter: The number of random iterations to calculate the approximated power of the Jacobian trace for each interest point.
        norm_weights: Whether to normalize the returned weights (to get values between 0 and 1).

    Returns: A list of (possibly normalized) jacobian-based weights to be considered as the relevancy that each interest
    point's output has on the model's output.
    """

    if not all([images.shape[0] == 1 for node, images in model_input_tensors.items()]):
        Logger.critical("Iterative jacobian trace computation is only supported on a single image sample")

    with tf.GradientTape(persistent=True, watch_accessed_variables=False) as g:
        outputs, interest_points_tensors = _model_outputs_computation(graph_float,
                                                                      model_input_tensors,
                                                                      interest_points,
                                                                      output_list,
                                                                      gradient_tape=g)
        outputs_jacobians_approx = []
        for output in outputs:  # Per model's output tensor
            output = tf.reshape(output, shape=[output.shape[0], -1])

            ipts_jac_trace_approx = []
            for ipt in interest_points_tensors:  # Per Interest point activation tensor
                trace_jv = []
                for j in range(n_iter):  # Approximation iterations
                    # Getting a random vector with normal distribution
                    v = tf.random.normal(shape=output.shape)
                    f_v = tf.reduce_sum(v * output)

                    with g.stop_recording():
                        # Computing the jacobian approximation by getting the gradient of (output * v)
                        jac_v = g.gradient(f_v, ipt, unconnected_gradients=tf.UnconnectedGradients.ZERO)
                        jac_v = tf.reshape(jac_v, [jac_v.shape[0], -1])
                        jac_trace_approx = tf.reduce_mean(tf.reduce_sum(tf.pow(jac_v, 2.0)))
                        trace_jv.append(jac_trace_approx)
                ipts_jac_trace_approx.append(2 * tf.reduce_mean(trace_jv) / output.shape[-1])  # Get averaged squared jacobian trace approximation
            outputs_jacobians_approx.append(ipts_jac_trace_approx)

        mean_per_point = tf.reduce_mean(outputs_jacobians_approx, axis=0)  # Get mean of jacobian approx of all model outputs
        if norm_weights:
            return _normalize_weights(mean_per_point, all_outputs_indices, alpha)
        else:
            return mean_per_point


def _model_outputs_computation(graph_float: common.Graph,
                               model_input_tensors: Dict[BaseNode, np.ndarray],
                               interest_points:  List[BaseNode],
                               output_list: List[BaseNode],
                               gradient_tape: tf.GradientTape) -> Tuple[List[tf.Tensor], List[tf.Tensor]]:
    """
    Computes the model's output according to the given graph representation on the given input,
    while recording necessary intermediate tensors for gradients computation.

    Args:
        graph_float: Graph to build its corresponding Keras model.
        model_input_tensors: A mapping between model input nodes to an input batch.
        interest_points: List of nodes which we want to get their feature map as output, to calculate distance metric.
        output_list: List of nodes that considered as model's output for the purpose of gradients computation.
        gradient_tape: A GradientTape object for recording necessary info for computing gradients.

    Returns: A list of output tensors and a list of activation tensors of all interest points.

    """

    node_to_output_tensors_dict = dict()

    # Build an OperationHandler to handle conversions from graph nodes to Keras operators.
    oh = OperationHandler(graph_float)
    input_nodes_to_input_tensors = {inode: tf.convert_to_tensor(model_input_tensors[inode]) for
                                    inode in graph_float.get_inputs()}  # Cast numpy array to tf.Tensor

    interest_points_tensors = []
    output_tensors = []
    for n in oh.node_sort:
        # Build a dictionary from node to its output tensors, by applying the layers sequentially.
        op_func = oh.get_node_op_function(n)  # Get node operation function

        input_tensors = build_input_tensors_list(n,
                                                 graph_float,
                                                 node_to_output_tensors_dict)  # Fetch Node inputs
        out_tensors_of_n = run_operation(n,  # Run node operation and fetch outputs
                                         input_tensors,
                                         op_func,
                                         input_nodes_to_input_tensors)

        # Gradients can be computed only on float32 tensors
        if isinstance(out_tensors_of_n, list):
            for i, t in enumerate(out_tensors_of_n):
                out_tensors_of_n[i] = tf.dtypes.cast(t, tf.float32)
        else:
            out_tensors_of_n = tf.dtypes.cast(out_tensors_of_n, tf.float32)

        if n in interest_points:
            # Recording the relevant feature maps onto the gradient tape
            gradient_tape.watch(out_tensors_of_n)
            interest_points_tensors.append(out_tensors_of_n)
        if n in output_list:
            output_tensors.append(out_tensors_of_n)

        if isinstance(out_tensors_of_n, list):
            node_to_output_tensors_dict.update({n: out_tensors_of_n})
        else:
            node_to_output_tensors_dict.update({n: [out_tensors_of_n]})

    return output_tensors, interest_points_tensors


def _normalize_weights(jacobians_traces: List,
                       all_outputs_indices: List[int],
                       alpha: float) -> List[float]:
    """
    Output layers or layers that come after the model's considered output layers,
    are assigned with a constant normalized value, according to the given alpha variable and the number of such layers.
    Other layers returned weights are normalized by dividing the jacobian-based weights value by the sum of all other values.

    Args:
        jacobians_traces: The approximated average jacobian-based weights of each interest point.
        all_outputs_indices: A list of indices of all nodes that consider outputs.
        alpha: A multiplication factor.

    Returns: Normalized list of jacobian-based weights (for each interest point).

    """

    sum_without_outputs = sum([jacobians_traces[i] for i in range(len(jacobians_traces)) if i not in all_outputs_indices])
    normalized_grads_weights = [_get_normalized_weight(grad, i, sum_without_outputs, all_outputs_indices, alpha)
                                for i, grad in enumerate(jacobians_traces)]

    return normalized_grads_weights


def _get_normalized_weight(grad: float,
                           i: int,
                           sum_without_outputs: float,
                           all_outputs_indices: List[int],
                           alpha: float) -> float:
    """
    Normalizes the node's gradient value. If it is an output or output replacement node than the normalized value is
    a constant, otherwise, it is normalized by dividing with the sum of all gradient values.

    Args:
        grad: The gradient value.
        i: The index of the node in the sorted interest points list.
        sum_without_outputs: The sum of all gradients of nodes that are not considered outputs.
        all_outputs_indices: A list of indices of all nodes that consider outputs.
        alpha: A multiplication factor.

    Returns: A normalized jacobian-based weights.

    """

    if i in all_outputs_indices:
        return alpha / len(all_outputs_indices)
    else:
        return ((1 - alpha) * grad / (sum_without_outputs + EPS)).numpy()



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from copy import copy

import logging
import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.python.util.object_identity import Reference as TFReference
from typing import List

from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.graph.edge import Edge
from model_compression_toolkit.core.common.graph.base_node import BaseNode
from model_compression_toolkit.core.keras.reader.common import is_node_a_model, is_node_an_input_layer
from model_compression_toolkit.core.keras.reader.connectivity_handler import ConnectivityHandler
from model_compression_toolkit.core.keras.reader.nested_model.nested_model_handler import merge_graphs

keras = tf.keras
layers = keras.layers


def get_edges(connectivity_handler: ConnectivityHandler) -> List[Edge]:
    """
    Creates a list of edges the model has. Each edge contains source and destination nodes, and indices that are
    the relative position of the tensor that connects those two nodes (in output tensors of the source node, and
    in input tensors of the destination node).
    Args:
        connectivity_handler: Connectivity handler containing all information about connections in graph.

    Returns:
        List of edges in the graph.
    """
    edges = []
    for src_node in connectivity_handler.get_nodes():  # go over all nodes in the graph
        for dst_node, src_index, dst_index in connectivity_handler.get_out_edges_params_list(
                src_node):  # get parameters of all out edges for a node
            edge = Edge(src_node,
                        dst_node,
                        src_index,
                        dst_index)  # create edge
            edges.append(edge)  # append edge to list
    return edges


def build_tensors_list(tensors_list) -> List[TFReference]:
    """
    Build list of tensor references from a list of tensors or a single tensor.
    Args:
        tensors_list: A tensor or a list of tensors.

    Returns:
        List of tensors' references.
    """
    return [tensors_list.ref()] if not isinstance(tensors_list, (list, tuple)) else [tensor.ref()
                                                                                     for
                                                                                     tensor in
                                                                                     tensors_list]

def build_connectivity_handler(model: Model) -> ConnectivityHandler:
    """
    Build a connectivity handler containing all information about connections in the model (nodes and
    interconnecting tensors).
    For a reused layer, several nodes may be generated, and can be distinguished by their names.
    Args:
        model: Keras model to build its connectivity handler.

    Returns:
        Connectivity handler containing all information about connections in graph.
    """
    connectivity_handler = ConnectivityHandler()
    for nodes in model._nodes_by_depth.values():
        for node in nodes:  # nodes_by depth values are lists (each list for a different depth)
            node_inputs = node.input_tensors if is_node_an_input_layer(node) else node.keras_inputs
            input_tensors = build_tensors_list(node_inputs)  # build input tensors of the node
            output_tensors = build_tensors_list(node.output_tensors)  # build output tensors of the node
            connectivity_handler.add_node(node,
                                          input_tensors,
                                          output_tensors)  # add the node and its connecting tensors to the
    return connectivity_handler


def build_graph(model: Model,
                connectivity_handler: ConnectivityHandler) -> Graph:
    """
    Given a Keras model, build and return an networkx MultiDiGraph containing all data (nodes, edges,
    inputs and outputs) representing that model.

    Args:
        model: Keras model to build its graph.
        connectivity_handler: Connectivity handler containing all information about connections in graph.

    Returns:
        Networkx MultiDiGraph representing the Keras model.

    """
    # connectivity handler contains info with keras nodes, but we build the model's graph with internal nodes so we
    # need to convert connectivity handler's data to internal nodes.
    connectivity_handler.convert_to_internal_nodes()
    inputs = connectivity_handler.build_inputs_list([t.ref() for t in model.inputs])
    outputs = connectivity_handler.build_outputs_list([t.ref() for t in model.outputs])
    edges = get_edges(connectivity_handler)
    nodes = connectivity_handler.get_nodes()
    return Graph(model.name, nodes, inputs, outputs, edges)


def parse_model(model: Model) -> Graph:
    """
    Parse a Keras model into a Graph.
    In case of a nested model, it recursively unrolls inner models.

    Args:
        model: Keras model to build its graph.

    Returns:
        Networkx MultiDiGraph representing the Keras model including: nodes, edges, inputs, and outputs.
    """
    connectivity_handler = build_connectivity_handler(model)
    model_graph = build_graph(model, connectivity_handler)

    # Go over all nodes in the graph, and if one of them is a model by itself, unroll it recursively by
    # merging the inner model's graph into the outer model's graph.
    nodes = copy(model_graph.nodes)
    for node in nodes:
        if is_node_a_model(node):  # if the node represents a Keras model - flat it recursively
            model_graph = flatten_nested_model(model_graph,
                                               node,
                                               model)
    return model_graph


def flatten_nested_model(outer_graph: Graph,
                         inner_model_node: BaseNode,
                         outer_keras_model: Model):
    """
    Flat a nested model given two graphs: inner and outer models' graphs.
    The resulting graph contains all nodes of the inner model (except for input nodes and the node representing the
    model itself). Also, outputs are updated if the inner model is an output of the outer model (by inner model's
    output nodes), and edges to the inner model node are rewired.

    Args:
        outer_graph: Networkx MultiDiGraph of outer model.
        inner_model_node: Node in outer_graph representing the inner model.
        outer_keras_model: Keras model for retrieving the inner Keras model.

    Returns:
        Networkx MultiDiGraph object representing the Keras model when inner model is flattened.
    """

    inner_model = outer_keras_model.get_layer(inner_model_node.name)  # get the Keras model
    inner_graph = parse_model(inner_model)  # recursively parse this model into graph.
    # merge graphs into a single graph representing the model when the inner model is being unrolled.
    outer_flatten_graph = merge_graphs(inner_model_node,
                                       outer_graph,
                                       inner_graph)
    return outer_flatten_graph


def model_reader(keras_model: Model) -> Graph:
    """
    Reads a Keras model and builds a base graph representing the model.
    Args:
        keras_model: Keras model to build its graph representation.

    Returns:
        Base graph of the Keras model.
    """
    logging.info("Start Model Reading...")
    graph = parse_model(keras_model)
    return graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
import torch
import torch.nn as nn
from typing import List, Union, Tuple, Dict
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.gptq.pytorch.quantizer.gumbel_rounding.base_gumbel_weights_quantizer import BaseGumbelWeightQuantizer, init_aux_var
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor, torch_tensor_to_numpy
from model_compression_toolkit.gptq.pytorch.quantizer.quant_utils import ste_clip, ste_gumbel, gumbel_softmax
from model_compression_toolkit.gptq.common.gptq_constants import AUXVAR, PTQ_MAX_RANGE, PTQ_MIN_RANGE, TEMP
from model_compression_toolkit.core.common.quantization.node_quantization_config import NodeWeightsQuantizationConfig
from model_compression_toolkit.core.common.constants import RANGE_MAX, RANGE_MIN

class UniformGumbelWeightQuantizer(BaseGumbelWeightQuantizer):
    """
    Class that implements a quantizer with trainable parameters to be used for GPTQ training.
    """

    def __init__(self,
                 weights_quantization_cfg: NodeWeightsQuantizationConfig,
                 gptq_config: GradientPTQConfig,
                 weight_shape: torch.Size):
        """
        Construct a Pytorch model that utilize a fake weight quantizer of Uniform Gumbel rounding
        Args:
            weights_quantization_cfg: Configuration of weight quantization
            gptq_config: GradientPTQConfig object with parameters about the tuning process.
            weight_shape: weight shape for auxiliary tensor creation.
        """
        super().__init__(weights_quantization_cfg, gptq_config, weight_shape)
        self.min_int = 0
        self.max_int = 2**self.num_bits - 1
        self.max_range_tensor = weights_quantization_cfg.weights_quantization_params.get(RANGE_MAX)
        self.min_range_tensor = weights_quantization_cfg.weights_quantization_params.get(RANGE_MIN)

        # Set trainable tensors
        self.set_trainable_params()


    def set_trainable_params(self):
        """
        A function to set a list of trainable parameters of the quantizer for GPTQ retraining
        """
        self.aux_tensor = nn.Parameter(to_torch_tensor(init_aux_var(self.weight_shape, self.m)), requires_grad=True)
        self.trainable_params.update({AUXVAR: self.aux_tensor})
        self.temp_tensor = nn.Parameter(to_torch_tensor(self.maximal_temp*torch.ones([1,*self.weight_shape])), requires_grad=True)
        self.trainable_params.update({TEMP: self.temp_tensor})
        self.max_range_tensor = nn.Parameter(to_torch_tensor(self.max_range_tensor), requires_grad=self.quantization_parameter_learning)
        self.trainable_params.update({PTQ_MAX_RANGE: self.max_range_tensor})
        self.min_range_tensor = nn.Parameter(to_torch_tensor(self.min_range_tensor), requires_grad=self.quantization_parameter_learning)
        self.trainable_params.update({PTQ_MIN_RANGE: self.min_range_tensor})

    def get_aux_variable(self) -> torch.Tensor:
        """
        Returns auxiliary trainable variables
        """
        return self.trainable_params.get(AUXVAR)

    def get_quantization_variable(self) -> Union[torch.Tensor, List]:
        """
        Returns quantization trainable variables
        """
        return [self.trainable_params.get(PTQ_MAX_RANGE), self.trainable_params.get(PTQ_MIN_RANGE)]

    def get_temperature_variable(self) -> Union[torch.Tensor, List]:
        """
        Returns temperature trainable variables
        """
        return self.trainable_params.get(TEMP)

    def get_weight_quant_params(self) -> Dict[str, np.ndarray]:
        """
        Returns weight quantization dictionary params
        """
        max_range_tensor = self.max_range_tensor
        min_range_tensor = self.min_range_tensor
        return {PTQ_MAX_RANGE: torch_tensor_to_numpy(max_range_tensor.detach()),
                PTQ_MIN_RANGE: torch_tensor_to_numpy(min_range_tensor.detach())}

    @staticmethod
    def fix_range_to_include_zero(range_min: torch.Tensor,
                                  range_max: torch.Tensor,
                                  n_bits: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Adjusting the quantization range to include representation of 0.0 in the quantization grid.
        If quantization per-channel, then range_min and range_max should be tensors in the specific shape that allows
        quantization along the channel_axis.

        Args:
            range_min: min bound of the quantization range (before adjustment).
            range_max: max bound of the quantization range (before adjustment).
            n_bits: Number of bits to quantize the tensor.

        Returns: adjusted quantization range
        """
        min_positive = range_min > 0
        max_negative = range_max < 0
        mid_range = torch.logical_and(torch.logical_not(min_positive), torch.logical_not(max_negative))
        min_positive = min_positive.float()
        max_negative = max_negative.float()
        mid_range = mid_range.float()

        scale = (range_max - range_min) / (2 ** n_bits - 1)
        min_range_adj = scale * torch.round(range_min / scale)
        max_range_adj = range_max - range_min + min_range_adj

        min_range_adj = min_range_adj * mid_range + max_negative * range_min
        max_range_adj = max_range_adj * mid_range + min_positive * range_max
        return min_range_adj, max_range_adj

    def forward(self, w: nn.Parameter, training:bool = True) -> nn.Parameter:
        """
        Weight fake quantizer
        Args:
            w: weights to quantize.
            training: whether in training mode or not
        Returns:
            quantized weights
        """
        self.update_iteration(training)

        #####################################################
        # Gumbel Softmax
        #####################################################
        if training:
            self.p_t = gumbel_softmax(self.aux_tensor, self.tau, self.g_t)
        else:
            self.p_t = ste_gumbel(gumbel_softmax(self.aux_tensor, self.minimal_temp, 0))

        auxhat_tensor = torch.sum(self.p_t * self.shift_tensor.reshape(self.reshape_aux_shift), dim=0)

        #####################################################
        # Quantizer
        #####################################################
        max_range_tensor = self.max_range_tensor
        min_range_tensor = self.min_range_tensor

        # adjusts the quantization rage so the quantization grid include zero.
        a, b = self.fix_range_to_include_zero(min_range_tensor, max_range_tensor, self.num_bits)

        # Compute the step size of quantized values.
        delta_tensor = (b - a) / (2 ** self.num_bits - 1)

        # Apply rounding
        w0 = torch.round((w - a) / delta_tensor).detach()  # Apply rounding

        w1 = w0 + auxhat_tensor

        # Clip data in range
        w2 = ste_clip(w1, min_val=self.min_int, max_val=self.max_int)

        # Quantize the data between min/max of quantization range.
        w_q = delta_tensor * w2 + a
        return w_q




# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from . import base_matcher
from . import edge_matcher
from . import node_matcher
from . import walk_matcher


def is_node_matcher(matcher: base_matcher.BaseMatcher):
    """
    Check whether a matcher is of type node matcher or not.
    Args:
        matcher: Matcher to check.

    Returns:
        Whether a matcher is of type node matcher or not.
    """
    return issubclass(matcher.__class__, node_matcher.BaseNodeMatcher)


def is_edge_matcher(matcher: base_matcher.BaseMatcher):
    """
    Check whether a matcher is of type edge matcher or not.
    Args:
        matcher: Matcher to check.

    Returns:
        Whether a matcher is of type edge matcher or not.
    """
    return issubclass(matcher.__class__, edge_matcher.BaseEdgeMatcher)


def is_walk_matcher(matcher: base_matcher.BaseMatcher):
    """
    Check whether a matcher is of type walk matcher or not.
    Args:
        matcher: Matcher to check.

    Returns:
        Whether a matcher is of type walk matcher or not.
    """
    return issubclass(matcher.__class__, walk_matcher.WalkMatcherList)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np

import model_compression_toolkit.core.common.quantization.quantization_config as qc
from model_compression_toolkit.core.common.constants import MIN_THRESHOLD, THRESHOLD
from model_compression_toolkit.core.common.quantization.quantization_params_generation.qparams_search import \
    qparams_selection_tensor_search, qparams_selection_histogram_search
from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import max_power_of_two, get_tensor_max
from model_compression_toolkit.core.common.quantization.quantization_params_generation.error_functions import \
    get_threshold_selection_tensor_error_function, get_threshold_selection_histogram_error_function
from model_compression_toolkit.core.common.target_platform import QuantizationMethod


def power_of_two_selection_tensor(tensor_data: np.ndarray,
                                  p: int,
                                  n_bits: int,
                                  per_channel: bool = False,
                                  channel_axis: int = 1,
                                  n_iter: int = 10,
                                  min_threshold: float = MIN_THRESHOLD,
                                  quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the power of two threshold based on the provided QuantizationErrorMethod to quantize the tensor.
    Different search is applied, depends on the value of the selected QuantizationErrorMethod.

    Args:
        tensor_data: Tensor content as Numpy array.
        p: p-norm to use for the Lp-norm distance.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the quantization should be per-channel or not.
        channel_axis: Output channel index.
        n_iter: Number of iterations to search for the optimal threshold (not used for this method).
        min_threshold: Minimal threshold to use if threshold is too small (not used for this method).
        quant_error_method: an error function to optimize the parameters' selection accordingly.

    Returns:
        Power of two threshold to quantize the tensor in a power of 2 manner.
    """

    if quant_error_method == qc.QuantizationErrorMethod.NOCLIPPING:
        tensor_data = np.abs(tensor_data)
        tensor_max = get_tensor_max(tensor_data, per_channel, channel_axis)
        threshold = max_power_of_two(tensor_max, min_threshold)
    else:
        signed = True  # weights are always signed
        error_function = get_threshold_selection_tensor_error_function(QuantizationMethod.POWER_OF_TWO,
                                                                       quant_error_method, p, norm=False, n_bits=n_bits,
                                                                       signed=signed)
        threshold = qparams_selection_tensor_search(error_function,
                                                    tensor_data,
                                                    n_bits,
                                                    per_channel=per_channel,
                                                    channel_axis=channel_axis,
                                                    n_iter=n_iter,
                                                    min_threshold=min_threshold,
                                                    signed=signed)
    return {THRESHOLD: threshold}


def power_of_two_selection_histogram(bins: np.ndarray,
                                     counts: np.ndarray,
                                     p: int,
                                     n_bits: int,
                                     min_value: float,
                                     max_value: float,
                                     constrained: bool = True,
                                     n_iter: int = 20,
                                     min_threshold: float = MIN_THRESHOLD,
                                     quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the power of two threshold based on the provided QuantizationErrorMethod to quantize a histogram.
    Different search is applied, depends on the value of the selected QuantizationErrorMethod.

    Args:
        bins: Bins values of the histogram.
        counts: Bins counts of the histogram.
        p: p-norm to use for the Lp-norm distance (used only for lp threshold selection).
        n_bits: Number of bits to quantize the tensor.
        min_value: Min value (not used for this method).
        max_value: Max value (not used for this method).
        constrained: Whether the threshold should be constrained or not (not used for this method).
        n_iter: Number of iteration ot search for the threshold (not used for this method).
        min_threshold: Minimal threshold to use if threshold is too small (used only for kl threshold selection).
        quant_error_method: an error function to optimize the parameters' selection accordingly.

    Returns:
        Power of two threshold to quantize the histogram a power of 2 manner.
    """
    if quant_error_method == qc.QuantizationErrorMethod.NOCLIPPING:
        tensor_max = np.max(np.abs(bins)[1:][counts > 0])
        threshold = max_power_of_two(tensor_max, min_threshold)
    else:
        error_function = get_threshold_selection_histogram_error_function(QuantizationMethod.POWER_OF_TWO,
                                                                          quant_error_method, p)
        threshold = qparams_selection_histogram_search(error_function,
                                                       bins,
                                                       counts,
                                                       n_bits,
                                                       constrained=constrained,
                                                       n_iter=n_iter,
                                                       min_threshold=min_threshold)
    return {THRESHOLD: threshold}


def power_of_two_no_clipping_selection_min_max(bins: np.ndarray,
                                               counts: np.ndarray,
                                               p: int,
                                               n_bits: int,
                                               min_value: float,
                                               max_value: float,
                                               constrained: bool = False,
                                               n_iter: int = 20,
                                               min_threshold: float = MIN_THRESHOLD,
                                               quant_error_method: qc.QuantizationErrorMethod =
                                               qc.QuantizationErrorMethod.NOCLIPPING) -> dict:
    """
    Gets a threshold between min and max numbers.
    If computed threshold is less than min_threshold, min_threshold is returned.

    Returns:
        A power of 2 threshold of the min/max values.
    """
    return power_of_two_selection_histogram(np.asarray([min_value, max_value]),  # histogram with min-max values only
                                            np.asarray([1]),  # passing dummy counts just to make the dimensions work
                                            p,
                                            n_bits,
                                            min_value,
                                            max_value,
                                            constrained,
                                            n_iter,
                                            min_threshold=min_threshold,
                                            quant_error_method=qc.QuantizationErrorMethod.NOCLIPPING)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List
import torch

def mse_loss(y: torch.Tensor, x: torch.Tensor, normalized: bool = True) -> torch.Tensor:
    """
    Compute the MSE of two tensors.
    Args:
        y: First tensor.
        x: Second tensor.
        normalized: either return normalized MSE or MSE
    Returns:
        The MSE of two tensors.
    """
    loss = torch.nn.MSELoss()(x,y)
    return loss / torch.mean(torch.square(x)) if normalized else loss


def multiple_tensors_mse_loss(y_list: List[torch.Tensor],
                              x_list: List[torch.Tensor],
                              fxp_w_list: List[List[torch.Tensor]],
                              flp_w_list: List[List[torch.Tensor]],
                              act_bn_mean: List,
                              act_bn_std: List,
                              loss_weights: torch.Tensor = None) -> torch.Tensor:
    """
    Compute MSE similarity between two lists of tensors

    Args:
        y_list: First list of tensors.
        x_list: Second list of tensors.
        fxp_w_list: list of lists each containing a quantized model layer's trainable weights - quantized
        flp_w_list: list of lists each containing a float model layer's weights - not quantized
        act_bn_mean: list of prior activations mean collected from batch normalization. None is there's no info
        act_bn_std: list of prior activations std collected from batch normalization. None is there's no info
        loss_weights: A vector of weights to compute weighted average loss.
    Returns:
        A single loss value which is the average of all MSE loss of all tensor pairs
        List of MSE similarities per tensor pair
    """

    loss_values_list = []
    for i, (y, x) in enumerate(zip(y_list, x_list)):
        point_loss = mse_loss(y, x)
        loss_values_list.append(point_loss)

    if loss_weights is not None:
        return torch.mean(loss_weights * torch.stack(loss_values_list))
    else:
        return torch.mean(torch.stack(loss_values_list))




# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Union
import torch
from torch.nn.functional import softmax, log_softmax, one_hot
from model_compression_toolkit.core.common.constants import MIN_THRESHOLD
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor


def power_of_two_max(max_tensor: torch.Tensor) -> torch.Tensor:
    """
    Compute the power of two threshold for a tensor.
    """
    return torch.pow(2, ste_ceil(torch.log2(torch.clip(max_tensor, min=MIN_THRESHOLD, max=torch.inf))))


def ste_ceil(x: torch.Tensor) -> torch.Tensor:
    """
    Return the ceil values of a tensor.
    """
    return (torch.ceil(x) - x).detach() + x


def ste_round(x: torch.Tensor) -> torch.Tensor:
    """
    Calculate the rounded values of a tensor
    Args:
        x: input variable
    Returns:
        rounded value
    """
    return (torch.round(x) - x).detach() + x


def ste_clip(x: torch.Tensor, min_val=-1.0, max_val=1.0) -> torch.Tensor:
    """
    Clip a variable between fixed values such that min_val<=output<=max_val
    Args:
        x: input variable
        min_val: minimum value for clipping
        max_val: maximum value for clipping
    Returns:
        clipped variable
    """
    return (torch.clip(x, min=min_val, max=max_val) - x).detach() + x


def gumbel_softmax(x: torch.Tensor, tau: Union[torch.Tensor,float], gumbel_tensor: Union[torch.Tensor,float], eps: float = 1e-6, axis=0) -> torch.Tensor:
    """
    A gumbel softmax function.
    Args:
        x: A tensor of log probability.
        tau: A temperature tensor.
        gumbel_tensor: A tensor of gumbel random variable.
        eps: A small number for numeric stability.
        axis: A integer representing the axis of which the gumbel softmax applyed on.

    Returns: A gumbel softmax probability tensor.

    """
    return softmax((log_softmax(x, dim=axis) + gumbel_tensor) / (tau + eps), dim=axis)


def select_gumbel(prob: torch.Tensor) -> torch.Tensor:
    """
    This function apply ste on the output of the gumbel softmax.
    Args:
        prob: A tensor of probability.

    Returns: A Tensor of ohe hot vector

    """
    max_index = torch.argmax(prob, dim=0)
    axis_list = [i for i in range(len(max_index.shape))]
    axis_list.insert(0, len(max_index.shape))
    one_hot_prob = torch.permute(one_hot(max_index, num_classes=prob.shape[0]), axis_list)
    return one_hot_prob + 0*prob


def ste_gumbel(prob: torch.Tensor) -> torch.Tensor:
    """
    This function apply ste on the output of the gumbel softmax.
    Args:
        prob:A tensor of probability

    Returns: A Tensor of ohe hot vector with STE.

    """
    delta = (select_gumbel(prob) - prob).detach()
    return prob + delta


def sample_gumbel(shape, eps=1e-6) -> torch.Tensor:
    """
    A function that sample a tensor of i.i.d gumbel random variable.
    Args:
        shape: The tensor output shape
        eps: A small number for numeric stability.

    Returns: A tensor of i.i.d gumbel random variable.

    """
    u = to_torch_tensor(torch.rand(shape))
    return -torch.log(-torch.log(u + eps) + eps)